function varargout=ekf(flag,varargin)
% *********************************************************************
% Implements an Extented Kalman filter.
% ---------------------------------------------------------------------
% Usage: [xht,St,yht,rest,Rt]=ekf(3,pred,yt,ut) % off line version
%                             ekf(3,pred,yt)    % off line version
%        [xh,P]=ekf(0,pred)           % Initialisering recursive version
%        [xh,P,res,yh,R,K]=ekf(1,y)   % Data update
%        [xh,P]=ekf(2,u)              % Time update
%        [xh,P]=ekf(2)                % Time update
% ---------------------------------------------------------------------
% yt    data sequence of measurements (outputs)
% ut    data sequence of inputs
% y     instant measurement
% u     instant input
% pred  predictive (1) or ordinary (0) version
% *********************************************************************
persistent stype
persistent pred
persistent nx nu nv ny ne
persistent x0 P0 vm R1 em R2
persistent xh P

switch flag,
 % *********************************************************************
 case 0, % Initialisering
  if nargin>1, pred=varargin{1}; else, pred=0; end;
  [stype,nx,nu,nv,ny,ne]=initt(3);
  [x0,P0,vm,R1,em,R2]=initt(4);
  xh=x0; P=P0;
  varargout={xh, P};
  % ------------------------------------------------------------------------
  % Data update
  % ------------------------------------------------------------------------
 case 1,
  y=varargin{1};
  yh=measmodel([xh; em]);
  [C,D,Gy]=measlin([xh; em]);
  if isempty(Gy),
   R=C*P*C'+R2;
  else
   R=C*P*C'+Gy*R2*Gy';
  end
  K=P*C'*inv(R);
  res=y-yh;
  if ~misobs(y),
   xh=xh+K*res;
   P=P-K*C*P;
  end;
  varargout={xh, P, res, yh, R, K};
  % -----------------------------------------------------------------------
  % Time update
  % ------------------------------------------------------------------------
 case 2
  if nu>0, u=varargin{1}; else, u=[]; end;
  [A,B,Gx]=statelin([xh; u; vm]);
  xh=statemodel([xh; u; vm]);
  if isempty(Gx),
   P=A*P*A'+R1;
  else
   P=A*P*A'+Gx*R1*Gx';
  end;
  varargout={xh,P};
  % *********************************************************************
  % * Offline version                                                   *
  % *********************************************************************
 case 3, 
  pred=varargin{1};
  ekf(0,pred);
  Yt=varargin{2};
  if nu>0, Ut=varargin{3}; end;

  xht=[]; St=[]; yht=[]; rest=[]; Rt=[];

  Ns=length(Yt);
  for i=1:Ns,
   fprintf(' %4i \r',i)
   if pred,
    xht=[xht; xh'];
    St=[St; sqrt(diag(P))'];
   end;

   y=Yt(:,i); % Data update
   [xh,P,res,yh,R,K]=ekf(1,y);
   rest=[rest; res'];
   yht=[yht; yh'];
   Rt=[Rt; sqrt(diag(R))'];

   if ~pred,
    xht=[xht; xh'];
    St=[St; sqrt(diag(P))'];
   end;
   if nu>0, % Time update
    u=Ut(:,i);
    [xh,P]=ekf(2,u);
   else,
    [xh,P]=ekf(2);
   end;
  end;
  xht=xht'; St=St'; yht=yht'; rest=rest'; Rt=Rt';
  varargout={xht; St; yht; rest; Rt};

  % *********************************************************************
 otherwise
  % *********************************************************************
end % flag end

