function varargout=ukf(flag,varargin)
% *********************************************************************
% Implements an Uncented Kalman filter.
% ---------------------------------------------------------------------
% Usage: [xht,St,yht,rest,Rt]=ukf(3,pred,yt,ut) % off line version
%                             ukf(3,pred,yt)    % off line version
%        [xh,P]=ukf(0,pred)           % Initialisering recursive version
%        [xh,P,res,yh,R,K]=ukf(1,y)   % Data update
%        [xh,P]=ukf(2,u)              % Time update
%        [xh,P]=ukf(2)                % Time update
% ---------------------------------------------------------------------
% yt    data sequence of measurements (outputs)
% ut    data sequence of inputs
% y     instant measurement
% u     instant input
% pred  predictive (1) or ordinary (0) version
% *********************************************************************
persistent pred kappa
persistent nx nu nv ny ne
persistent x0 P0 vm R1 em R2
persistent xh P
persistent Sv Se
persistent xht St Rt yht rest
persistent itime

switch flag,
 % *********************************************************************
 case 0, % Initialisering
  if nargin>1, pred=varargin{1}; else, pred=0; end;
  kappa=0.5;
  [stype,nx,nu,nv,ny,ne]=initt(3);
  [x0,P0,vm,R1,em,R2]=initt(4);
  Sv=msqrt((nx+nv+kappa)*R1);
  Se=msqrt((nx+ne+kappa)*R2);
  xh=x0; P=P0;
  varargout={xh, P};
  % ------------------------------------------------------------------------
  % Data update
  % ------------------------------------------------------------------------
 case 1,
  y=varargin{1};
  nz=nx+ne;
  Sx=msqrt((nz+kappa)*P);
  xa=kron(ones(1,nx),xh); xb=kron(ones(1,ne),xh);
  ea=kron(ones(1,ne),em); eb=kron(ones(1,nx),em);
  X=[xh xa+Sx xb xa-Sx xb];
  E=[em eb ea+Se eb ea-Se];
  w=[kappa/(nz+kappa) kron(ones(1,2*nz),1/2/(nz+kappa))];
  Y=measmodel([X; E]);

  yh=kron(ones(ny,1),w).*Y;
  yh=sum(yh')';
  Dy=Y-kron(ones(1,size(Y,2)),yh);
  R=Dy*diag(w)*Dy';

  Dx=X-kron(ones(1,size(X,2)),xh);
  Pxy=Dx*diag(w)*Dy';
  K=Pxy*inv(R);
  res=y-yh;
  if ~misobs(y),
   xh=xh+K*res;
   P=P-K*Pxy';
  end;
  varargout={xh, P, res, yh, R, K};
  % ------------------------------------------------------------------------
  % time update
  % ------------------------------------------------------------------------
 case 2,
  if nu>0, u=varargin{1}; else, u=[]; end;
  nz=nx+nv;
  Sx=msqrt((nz+kappa)*P);
  xa=kron(ones(1,nx),xh); xb=kron(ones(1,nv),xh);
  va=kron(ones(1,nv),vm); vb=kron(ones(1,nx),vm);
  X=[xh xa+Sx xb xa-Sx xb];
  if nu>0, U=kron(ones(1,1+2*(nx+nv)),u); else, U=[]; end;
  V=[vm vb va+Sv vb va-Sv];
  w=[kappa/(nz+kappa) kron(ones(1,2*nz),1/2/(nz+kappa))];
  X=statemodel([X; U; V]);

  xh=kron(ones(nx,1),w).*X;
  xh=sum(xh')';

  Dx=X-kron(ones(1,size(X,2)),xh);
  P=Dx*diag(w)*Dx';
  varargout={xh,P};

  % *********************************************************************
 otherwise
  % *********************************************************************
end % flag end
